@@ -55,9 +55,8 @@ unsigned int blocksize = DEFAULT_BLOCKSIZE;
 
 int run_dedupe = 0;
 int recurse_dirs = 0;
-int one_file_system = 1;
-int block_dedupe = 0;
-int dedupe_same_file = 0;
+int v2_hashfile = 0;
+int dedupe_same_file = 1;
 int skip_zeroes = 0;
 
 int target_rw = 1;
@@ -67,20 +66,23 @@ static int fdupes_mode = 0;
 static int stdin_filelist = 0;
 static unsigned int list_only_opt = 0;
 static unsigned int rm_only_opt = 0;
+struct dbfile_config dbfile_cfg;
+static bool force_v2_hashfile = false;
+static int partial_extent_search = 0;
 
 static enum {
 	H_READ,
 	H_WRITE,
 	H_UPDATE,
 } use_hashfile = H_UPDATE;
-static char *serialize_fname = NULL;
+char *serialize_fname = NULL;
 static unsigned int nr_logical_cpus;
 static unsigned int nr_physical_cpus;
 unsigned int io_threads;
 unsigned int cpu_threads;
 int io_threads_opt = 0;
 int cpu_threads_opt = 0;
-int do_lookup_extents = 0;
+int do_lookup_extents = 1;
 int fiemap_during_dedupe = 1;
 
 int stdout_is_tty = 0;
@@ -99,7 +101,7 @@ static int list_db_files(char *filename)
 {
 	int ret;
 
-	ret = dbfile_open(filename);
+	ret = dbfile_open(filename, &dbfile_cfg);
 	if (ret) {
 		fprintf(stderr, "Error: Could not open \"%s\"\n", filename);
 		return ret;
@@ -166,7 +168,7 @@ static int rm_db_files(char *dbfilename)
 	int ret, err = 0;
 	struct rm_file *rm, *tmp;
 
-	ret = dbfile_open(dbfilename);
+	ret = dbfile_open(dbfilename, &dbfile_cfg);
 	if (ret) {
 		fprintf(stderr, "Error: Could not open \"%s\"\n", dbfilename);
 		return ret;
@@ -184,7 +186,8 @@ restart:
 			 */
 			goto restart;
 		}
-		ret = dbfile_remove_file(dbfile_get_handle(), rm->filename);
+		ret = dbfile_remove_file(dbfile_get_handle(), &dbfile_cfg,
+					 rm->filename);
 		if (ret == 0)
 			vprintf("Removed \"%s\" from hashfile.\n",
 				rm->filename);
@@ -199,6 +202,17 @@ restart:
 	return err;
 }
 
+extern struct list_head exclude_list;
+
+static void add_exclude_pattern(const char *pattern)
+{
+	struct exclude_file *exclude = malloc(sizeof(*exclude));
+	if (exclude) {
+		exclude->pattern = strdup(pattern);
+		list_add_tail(&exclude->list, &exclude_list);
+	}
+}
+
 static void usage(const char *prog)
 {
 	char *s = NULL;
@@ -264,9 +278,11 @@ static int parse_dedupe_opts(const char *opts)
 		if (strcmp(token, "same") == 0) {
 			dedupe_same_file = !invert;
 		} else if (strcmp(token, "block") == 0) {
-			block_dedupe = !invert;
+			; /* This option ignored as of v0.12 */
 		} else if (strcmp(token, "fiemap") == 0) {
 			fiemap_during_dedupe = !invert;
+		} else if (strcmp(token, "partial") == 0) {
+			partial_extent_search = !invert;
 		} else {
 			print_usage = 1;
 			break;
@@ -290,17 +306,18 @@ enum {
 	HELP_OPTION,
 	VERSION_OPTION,
 	WRITE_HASHES_OPTION,
+	WRITE_OLD_HASHES_OPTION,
 	READ_HASHES_OPTION,
 	HASHFILE_OPTION,
 	IO_THREADS_OPTION,
 	CPU_THREADS_OPTION,
 	LOOKUP_EXTENTS_OPTION,
-	ONE_FILESYSTEM_OPTION,
 	HASH_OPTION,
 	SKIP_ZEROES_OPTION,
 	FDUPES_OPTION,
 	DEDUPE_OPTS_OPTION,
 	QUIET_OPTION,
+	EXCLUDE_OPTION,
 };
 
 static int add_files_from_stdin(int fdupes)
@@ -330,7 +347,7 @@ static int add_files_from_stdin(int fdupes)
 			continue;
 		}
 
-		if (add_file(path, AT_FDCWD))
+		if (add_file(path))
 			return 1;
 
 		/* Give the user a chance to see some output from add_file(). */
@@ -351,7 +368,7 @@ static int add_files_from_cmdline(int numfiles, char **files)
 	for (i = 0; i < numfiles; i++) {
 		const char *name = files[i];
 
-		if (add_file(name, AT_FDCWD))
+		if (add_file(name))
 			return 1;
 	}
 
@@ -373,25 +390,26 @@ static int parse_options(int argc, char **argv, int *filelist_idx)
 		{ "help", 0, NULL, HELP_OPTION },
 		{ "version", 0, NULL, VERSION_OPTION },
 		{ "write-hashes", 1, NULL, WRITE_HASHES_OPTION },
+		{ "write-hashes-v2", 1, NULL, WRITE_OLD_HASHES_OPTION },
 		{ "read-hashes", 1, NULL, READ_HASHES_OPTION },
 		{ "hashfile", 1, NULL, HASHFILE_OPTION },
 		{ "io-threads", 1, NULL, IO_THREADS_OPTION },
 		{ "hash-threads", 1, NULL, IO_THREADS_OPTION },
 		{ "cpu-threads", 1, NULL, CPU_THREADS_OPTION },
 		{ "lookup-extents", 1, NULL, LOOKUP_EXTENTS_OPTION },
-		{ "one-file-system", 0, NULL, ONE_FILESYSTEM_OPTION },
 		{ "hash", 1, NULL, HASH_OPTION },
 		{ "skip-zeroes", 0, NULL, SKIP_ZEROES_OPTION },
 		{ "fdupes", 0, NULL, FDUPES_OPTION },
 		{ "dedupe-options=", 1, NULL, DEDUPE_OPTS_OPTION },
 		{ "quiet", 0, NULL, QUIET_OPTION },
+		{ "exclude", 1, NULL, EXCLUDE_OPTION },
 		{ NULL, 0, NULL, 0}
 	};
 
 	if (argc < 2)
 		return 1;
 
-	while ((c = getopt_long(argc, argv, "Ab:vdDrh?xLR:q", long_ops, NULL))
+	while ((c = getopt_long(argc, argv, "Ab:vdDrh?LR:q", long_ops, NULL))
 	       != -1) {
 		switch (c) {
 		case 'A':
@@ -425,6 +443,8 @@ static int parse_options(int argc, char **argv, int *filelist_idx)
 		case 'h':
 			human_readable = 1;
 			break;
+		case WRITE_OLD_HASHES_OPTION:
+			force_v2_hashfile = true;
 		case WRITE_HASHES_OPTION:
 			write_hashes = 1;
 			serialize_fname = strdup(optarg);
@@ -458,10 +478,6 @@ static int parse_options(int argc, char **argv, int *filelist_idx)
 		case LOOKUP_EXTENTS_OPTION:
 			do_lookup_extents = parse_yesno_option(optarg, 0);
 			break;
-		case ONE_FILESYSTEM_OPTION:
-		case 'x':
-			one_file_system = 1;
-			break;
 		case HASH_OPTION:
 			user_hash = optarg;
 			break;
@@ -486,6 +502,9 @@ static int parse_options(int argc, char **argv, int *filelist_idx)
 		case 'q':
 			quiet = 1;
 			break;
+		case EXCLUDE_OPTION:
+			add_exclude_pattern(optarg);
+			break;
 		case HELP_OPTION:
 			help_option = 1;
 		case '?':
@@ -495,6 +514,9 @@ static int parse_options(int argc, char **argv, int *filelist_idx)
 		}
 	}
 
+	if (!do_lookup_extents || !fiemap_during_dedupe)
+		force_v2_hashfile = true;
+
 	numfiles = argc - optind;
 
 	/* Filter out option combinations that don't make sense. */
@@ -579,45 +601,57 @@ static void print_header(void)
 	qprintf("Gathering file list...\n");
 }
 
+static int update_config_from_dbfile(void)
+{
+	dedupe_seq = dbfile_cfg.dedupe_seq;
+
+	if (strncasecmp(dbfile_cfg.hash_type, hash_type, 8)) {
+		fprintf(stderr,
+			"Error: Hashfile %s uses %.*s. for checksums "
+			"but we are using %.*s.\nTry running with "
+			"--hash=%.*s\n", serialize_fname, 8,
+			dbfile_cfg.hash_type, 8, hash_type,
+			8, dbfile_cfg.hash_type);
+		return EINVAL;
+	}
+
+	if (dbfile_cfg.blocksize != blocksize &&
+	    dbfile_cfg.major == BLOCK_DEDUPE_DBFILE_VER) {
+		vprintf("Using blocksize %uK from hashfile (%uK "
+			"blocksize requested).\n", dbfile_cfg.blocksize/1024,
+			blocksize/1024);
+		blocksize = dbfile_cfg.blocksize;
+	}
+
+	return 0;
+}
+
 static int create_update_hashfile(int argc, char **argv, int filelist_idx)
 {
 	int ret;
 	int dbfile_is_new = 0;
 
-	ret = dbfile_create(serialize_fname, &dbfile_is_new);
+	ret = dbfile_create(serialize_fname, &dbfile_is_new,
+			    force_v2_hashfile ? BLOCK_DEDUPE_DBFILE_VER : DB_FILE_MAJOR,
+			    &dbfile_cfg);
 	if (ret)
 		goto out;
 
-	if (!dbfile_is_new) {
-		dev_t dev;
-		uint64_t fsid;
-		unsigned db_blocksize;
-		char db_hash_type[8];
+	if (force_v2_hashfile && dbfile_cfg.major != BLOCK_DEDUPE_DBFILE_VER) {
+		ret = EINVAL;
+		fprintf(stderr, "Error: asked to force hashfile version 2 but "
+			"existing hashfile has version %d\n", dbfile_cfg.major);
+		goto out;
+	}
 
-		ret = dbfile_get_config(&db_blocksize, NULL, NULL, &dev, &fsid,
-					NULL, NULL, db_hash_type, &dedupe_seq);
+	if (dbfile_cfg.major == BLOCK_DEDUPE_DBFILE_VER)
+		v2_hashfile = 1;
+
+	if (!dbfile_is_new) {
+		ret = update_config_from_dbfile();
 		if (ret)
 			goto out;
-
-		if (strncasecmp(db_hash_type, hash_type, 8)) {
-			fprintf(stderr,
-				"Error: Hashfile %s uses %.*s. for checksums "
-				"but we are using %.*s.\nTry running with "
-				"--hash=%.*s\n", serialize_fname, 8,
-				db_hash_type, 8, hash_type,
-				8, db_hash_type);
-			ret = EINVAL;
-			goto out;
-		}
-
-		if (db_blocksize != blocksize) {
-			vprintf("Using blocksize %uK from hashfile (%uK "
-				"blocksize requested).\n", db_blocksize/1024,
-				blocksize/1024);
-			blocksize = db_blocksize;
-		}
-
-		fs_set_onefs(dev, fsid);
+		fs_set_onefs(dbfile_cfg.onefs_dev, dbfile_cfg.onefs_fsid);
 	}
 
 	print_header();
@@ -631,14 +665,17 @@ static int create_update_hashfile(int argc, char **argv, int filelist_idx)
 		goto out;
 
 	if (dbfile_is_new) {
-		ret = dbfile_sync_config(blocksize, fs_onefs_dev(),
-					 fs_onefs_id(), dedupe_seq);
+		dbfile_cfg.blocksize = blocksize;
+		dbfile_cfg.onefs_dev = fs_onefs_dev();
+		dbfile_cfg.onefs_fsid = fs_onefs_id();
+		dbfile_cfg.dedupe_seq = dedupe_seq;
+		ret = dbfile_sync_config(&dbfile_cfg);
 		if (ret)
 			goto out;
 	} else {
 		qprintf("Adding files from database for hashing.\n");
 
-		ret = dbfile_scan_files();
+		ret = dbfile_scan_files(&dbfile_cfg);
 		if (ret)
 			goto out;
 	}
@@ -649,13 +686,13 @@ static int create_update_hashfile(int argc, char **argv, int filelist_idx)
 		goto out;
 	}
 
-	ret = populate_tree();
+	ret = populate_tree(&dbfile_cfg);
 	if (ret) {
 		fprintf(stderr,	"Error while populating extent tree!\n");
 		goto out;
 	}
 
-	ret = create_indexes(dbfile_get_handle());
+	ret = create_indexes(dbfile_get_handle(), &dbfile_cfg);
 	if (ret)
 		goto out;
 
@@ -724,6 +761,12 @@ int main(int argc, char **argv)
 	switch (use_hashfile) {
 	case H_UPDATE:
 	case H_WRITE:
+		if (serialize_fname)
+			printf("Warning: The hash file format in Duperemove "
+			       "master branch is under development and may "
+			       "change.\nIf the changes are not backwards "
+			       "compatible, you will have to re-create your "
+			       "hash file.\n");
 		ret = create_update_hashfile(argc, argv, filelist_idx);
 		if (ret)
 			goto out;
@@ -739,23 +782,25 @@ int main(int argc, char **argv)
 		}
 		break;
 	case H_READ:
-		ret = dbfile_open(serialize_fname);
+		ret = dbfile_open(serialize_fname, &dbfile_cfg);
 		if (ret) {
 			fprintf(stderr, "Error: Could not open dbfile %s.\n",
 				serialize_fname);
 			goto out;
 		}
 
+		if (dbfile_cfg.major == BLOCK_DEDUPE_DBFILE_VER)
+			v2_hashfile = 1;
+
 		/*
 		 * Skips the file scan, used to isolate the
 		 * extent-find and dedupe stages
 		 */
-		ret = dbfile_get_config(&blocksize, NULL, NULL, NULL, NULL,
-					NULL, NULL, NULL, &dedupe_seq);
-		if (ret) {
-			fprintf(stderr, "Error: initializing dbfile config\n");
+		blocksize = dbfile_cfg.blocksize;
+		ret = update_config_from_dbfile();
+		if (ret)
 			goto out;
-		}
+
 		print_header();
 		break;
 	default:
@@ -765,23 +810,41 @@ int main(int argc, char **argv)
 
 	qprintf("Loading only duplicated hashes from hashfile.\n");
 
-	ret = dbfile_load_hashes(&dups_tree);
-	if (ret)
-		goto out;
+	if (v2_hashfile) {
+		ret = dbfile_load_block_hashes(&dups_tree);
+		if (ret)
+			goto out;
 
-	ret = find_all_dupes(&dups_tree, &res);
-	if (ret) {
-		/* Only error for this should be enomem */
-		fprintf(stderr,
-			"Error %d: %s while finding duplicate extents.\n",
-			ret, strerror(ret));
-		goto out;
+		ret = find_all_dupes(&dups_tree, &res);
+		if (ret) {
+			/* Only error for this should be enomem */
+			fprintf(stderr,
+				"Error %d: %s while finding duplicate extents.\n",
+				ret, strerror(ret));
+			goto out;
+		}
+	} else {
+		ret = dbfile_load_extent_hashes(&res);
+		if (ret)
+			goto out;
+
+		ret = dbfile_load_block_hashes(&dups_tree);
+		if (ret)
+			goto out;
+
+		printf("Found %llu identical extents.\n", res.num_extents);
+		if (partial_extent_search) {
+			ret = find_additional_dedupe(&dups_tree, &res);
+			if (ret)
+				goto out;
+		}
 	}
 
-	if (run_dedupe) {
 #ifdef	PRINT_STATS
-		run_filerec_stats();
+	run_filerec_stats();
 #endif
+
+	if (run_dedupe) {
 		dedupe_results(&res, &dups_tree);
 
 		/*
@@ -791,18 +854,15 @@ int main(int argc, char **argv)
 		dedupe_seq++;
 
 		/* Sync to get new dedupe_seq written. */
-		ret = dbfile_sync_config(blocksize, fs_onefs_dev(),
-					 fs_onefs_id(), dedupe_seq);
+		dbfile_cfg.dedupe_seq = dedupe_seq;
+		dbfile_cfg.blocksize = blocksize;
+		dbfile_cfg.onefs_dev = fs_onefs_dev();
+		dbfile_cfg.onefs_fsid = fs_onefs_id();
+		ret = dbfile_sync_config(&dbfile_cfg);
 		if (ret)
 			goto out;
 	} else {
-		if (block_dedupe)
-			debug_print_hash_tree(&dups_tree);
-		else
-			print_dupes_table(&res);
-#ifdef	PRINT_STATS
-		run_filerec_stats();
-#endif
+		print_dupes_table(&res);
 	}
 
 out:
