@@ -19,6 +19,7 @@
 
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/ioctl.h>
 #include <limits.h>
 #include <fcntl.h>
 #include <assert.h>
@@ -30,9 +31,11 @@
 #include <string.h>
 #include <linux/limits.h>
 #include <linux/fiemap.h>
+#include <linux/fs.h>
 #include <inttypes.h>
 #include <linux/magic.h>
 #include <sys/statfs.h>
+#include <fnmatch.h>
 
 #include <glib.h>
 
@@ -62,11 +65,19 @@ static unsigned long long files_to_scan;
 static GMutex io_mutex; /* locks db writes */
 static unsigned int leading_spaces;
 
+#define READ_BUF_LEN (8*1024*1024) // 16m
+
 struct thread_params {
+	GMutex mutex;
 	int num_files;           /* Total number of files we hashed */
 	int num_hashes;          /* Total number of hashes we hashed */
+	struct dbfile_config	*dbfile_cfg; /* global dbfile config */
 };
 
+extern int v2_hashfile;
+extern struct dbfile_config dbfile_cfg;
+LIST_HEAD(exclude_list);
+
 static void set_filerec_scan_flags(struct filerec *file)
 {
 	if (!(file->flags & FILEREC_NEEDS_SCAN)) {
@@ -88,7 +99,6 @@ static void clear_filerec_scan_flags(struct filerec *file)
 void fs_set_onefs(dev_t dev, uint64_t fsid)
 {
 	if (dev || fsid) {
-		one_file_system = 1;
 		if (dev)
 			one_fs_dev = dev;
 		else if (fsid)
@@ -119,7 +129,7 @@ static int get_dirent_type(struct dirent *entry, int fd)
 	 * fashioned way. We translate mode to DT_* for the
 	 * convenience of the caller.
 	 */
-	ret = fstatat(fd, entry->d_name, &st, 0);
+	ret = fstatat(fd, entry->d_name, &st, AT_SYMLINK_NOFOLLOW);
 	if (ret) {
 		fprintf(stderr,
 			"Error %d: %s while getting type of file %s/%s. "
@@ -146,6 +156,21 @@ static int get_dirent_type(struct dirent *entry, int fd)
 	return DT_UNKNOWN;
 }
 
+static int is_excluded(const char *name)
+{
+	struct exclude_file *exclude, *tmp;
+
+	list_for_each_entry_safe(exclude, tmp, &exclude_list, list) {
+		if (fnmatch(exclude->pattern, name, FNM_PATHNAME) == 0) {
+			vprintf("Excluding: %s (matches %s)\n", name,
+				exclude->pattern);
+			return 1;
+		}
+	}
+
+	return 0;
+}
+
 static int walk_dir(const char *name)
 {
 	int ret = 0;
@@ -171,7 +196,7 @@ static int walk_dir(const char *name)
 			type = get_dirent_type(entry, dirfd(dirp));
 			if (type == DT_REG ||
 			    (recurse_dirs && type == DT_DIR)) {
-				if (add_file(entry->d_name, dirfd(dirp))) {
+				if (add_file(entry->d_name)) {
 					ret = 1;
 					goto out;
 				}
@@ -235,6 +260,9 @@ static int __add_file(const char *name, struct stat *st,
 		goto out;
 	}
 
+	if (is_excluded(name))
+		goto out;
+
 	if (run_dedupe &&
 	    ((fs.f_type != BTRFS_SUPER_MAGIC &&
 	      fs.f_type != XFS_SB_MAGIC))) {
@@ -282,18 +310,34 @@ out:
 	return 0;
 }
 
+static bool will_cross_mountpoint(dev_t dev, uint64_t btrfs_fsid)
+{
+	abort_on(one_fs_dev && one_fs_btrfs);
+
+	if (!one_fs_dev && !one_fs_btrfs) {
+		if (btrfs_fsid)
+			one_fs_btrfs = btrfs_fsid;
+		else
+			one_fs_dev = dev;
+	}
+
+	if ((one_fs_dev && (one_fs_dev != dev)) ||
+	    (one_fs_btrfs && (btrfs_fsid != one_fs_btrfs)))
+		return true;
+
+	return false;
+}
+
 /*
  * Returns nonzero on fatal errors only
  */
-int add_file(const char *name, int dirfd)
+int add_file(const char *name)
 {
 	int ret, len = strlen(name);
 	struct stat st;
 	char *pathtmp;
-	dev_t dev;
 	struct filerec *file = NULL;
 	char abspath[PATH_MAX];
-	uint64_t btrfs_fsid;
 
 	if (len > (path_max - pathp)) {
 		fprintf(stderr, "Path max exceeded: %s %s\n", path, name);
@@ -323,7 +367,7 @@ int add_file(const char *name, int dirfd)
 		goto out;
 	}
 
-	ret = stat(abspath, &st);
+	ret = lstat(abspath, &st);
 	if (ret) {
 		fprintf(stderr, "Error %d: %s while stating file %s. "
 			"Skipping.\n",
@@ -332,7 +376,12 @@ int add_file(const char *name, int dirfd)
 	}
 
 	if (S_ISDIR(st.st_mode)) {
-		dev = st.st_dev;
+		uint64_t btrfs_fsid;
+		dev_t dev = st.st_dev;
+
+		if (is_excluded(abspath))
+			goto out;
+
 		/*
 		 * Device doesn't work for btrfs as it changes between
 		 * subvolumes. We know how to get a unique fsid though
@@ -345,22 +394,11 @@ int add_file(const char *name, int dirfd)
 			goto out;
 		}
 
-		if (one_file_system) {
-			abort_on(one_fs_dev && one_fs_btrfs);
-
-			if (!one_fs_dev && !one_fs_btrfs) {
-				if (btrfs_fsid)
-					one_fs_btrfs = btrfs_fsid;
-				else
-					one_fs_dev = dev;
-			}
-
-			if ((one_fs_dev && (one_fs_dev != dev)) ||
-			    (one_fs_btrfs && (btrfs_fsid != one_fs_btrfs))) {
-				vprintf("Skipping file %s because of -x\n",
-					abspath);
-				goto out;
-			}
+		/* Don't cross mount points since dedup doesn't work across */
+		if (will_cross_mountpoint(dev, btrfs_fsid)) {
+			vprintf("Mountpoint traversal disallowed: %s \n",
+				abspath);
+			goto out;
 		}
 
 		if (walk_dir(name))
@@ -448,7 +486,7 @@ int add_file_db(const char *filename, uint64_t inum, uint64_t subvolid,
 			return 0;
 		}
 		/* Go to disk and look up by filename */
-		ret = stat(filename, &st);
+		ret = lstat(filename, &st);
 		if (ret == -1 && errno == ENOENT) {
 			vprintf("File path %s no longer exists. Skipping.\n",
 				filename);
@@ -499,26 +537,18 @@ int add_file_db(const char *filename, uint64_t inum, uint64_t subvolid,
 	return 0;
 }
 
-static GThreadPool *setup_pool(void *location, GMutex *mutex,
-			void *function)
+static GThreadPool *setup_pool(void *arg, void *function)
 {
 	GError *err = NULL;
 	GThreadPool *pool;
 
-	g_mutex_init(mutex);
-	g_dataset_set_data_full(location, "mutex", mutex,
-				(GDestroyNotify) g_mutex_clear);
-
-	pool = g_thread_pool_new((GFunc) function, location, io_threads,
-				 FALSE, &err);
+	pool = g_thread_pool_new((GFunc) function, arg, io_threads, FALSE,
+				 &err);
 	if (err != NULL) {
-		fprintf(
-			stderr,
-			"Unable to create thread pool: %s\n",
+		fprintf(stderr, "Unable to create thread pool: %s\n",
 			err->message);
 		g_error_free(err);
 		err = NULL;
-		g_dataset_destroy(location);
 		return NULL;
 	}
 	return pool;
@@ -549,116 +579,174 @@ static void run_pool(GThreadPool *pool)
 
 static inline int is_block_zeroed(void *buf, ssize_t buf_size)
 {
-	/*
-	 * If buf is block aligned check for zeroes may be accelerated
-	 * By checks block by CPU word size
-	 */
-	if (buf_size%sizeof(ssize_t) == 0) {
+	/* First check for zeroes in word-sized chunks */
+	if (buf_size >= sizeof(ssize_t)) {
 		ssize_t *buf_start = buf;
-		ssize_t *buf_end = buf+buf_size;
+		ssize_t *buf_end = buf + (buf_size - (buf_size % sizeof(ssize_t)));
 		ssize_t *ptr = buf_start;
 		for (; ptr < buf_end; ptr++) {
 			if (*ptr != 0)
 				return 0;
 		}
-	} else {
-		char *buf_start = buf;
-		char *buf_end = buf+buf_size;
-		char *ptr = buf_start;
-		for (; ptr < buf_end; ptr++) {
-			if (*ptr != 0)
-				return 0;
-		}
+
+		buf_size = buf_size % sizeof(ssize_t);
+		buf = buf_end;
+	}
+
+	char *buf_start = buf;
+	char *buf_end = buf+buf_size;
+	char *ptr = buf_start;
+	for (; ptr < buf_end; ptr++) {
+		if (*ptr != 0)
+			return 0;
 	}
 	return 1;
 }
 
-struct csum_block {
-	ssize_t bytes;
-	unsigned int flags;
+static int xlate_extent_flags(int fieflags, ssize_t len)
+{
+	int flags = 0;
+
+	if (fieflags & FIEMAP_SKIP_FLAGS)
+		flags |= FILE_BLOCK_SKIP_COMPARE;
+
+	if (len < blocksize)
+		flags |= FILE_BLOCK_PARTIAL;
+
+	return flags;
+}
+
+static int add_block_hash(struct block_csum **hashes, int *nr_hashes,
+			  uint64_t loff, unsigned char *digest, int flags)
+{
+	void *retp;
+	struct block_csum *block_hashes;
+
+	retp = realloc(*hashes, sizeof(struct block_csum) * (*nr_hashes + 1));
+	if (!retp)
+		return -ENOMEM;
+
+	block_hashes = retp;
+	block_hashes[*nr_hashes].loff = loff;
+	block_hashes[*nr_hashes].flags = flags;
+	memcpy(block_hashes[*nr_hashes].digest, digest, DIGEST_LEN_MAX);
+
+	*hashes = retp;
+	(*nr_hashes)++;
+	return 0;
+}
+
+struct csum_ctxt {
+	uint64_t blocks_recorded;
 	char *buf;
 	struct filerec *file;
 	unsigned char digest[DIGEST_LEN_MAX];
+
+	struct block_csum *block_hashes;
+	int nr_block_hashes;
+	unsigned char block_digest[DIGEST_LEN_MAX];
 };
 
-static inline int csum_next_block(struct csum_block *data, uint64_t *off,
-				  struct fiemap_ctxt **in_fc)
+static int csum_blocks(struct csum_ctxt *data, struct running_checksum *csum,
+		       const uint64_t extoff, const ssize_t extlen, int flags)
 {
-	struct fiemap_ctxt *fc = NULL;
-	ssize_t stored_bytes = data->bytes;
-	ssize_t bytes_read;
 	int ret = 0;
-	unsigned int hole;
-	int partial = 0;
+	int start = 0;
+	ssize_t cmp_len = extlen - start;
 
-	if (in_fc)
-		fc = *in_fc;
+	if (cmp_len > blocksize)
+		cmp_len = blocksize;
 
-	bytes_read = read(data->file->fd, data->buf + stored_bytes,
-				blocksize - stored_bytes);
-	if (bytes_read < 0) {
-		ret = errno;
-		fprintf(stderr, "Unable to read file %s: %s\n",
-			data->file->filename, strerror(ret));
-		return -1;
-	}
+	while (start < extlen) {
+		char *buf = data->buf + start;
 
-	/* Handle EOF */
-	if (bytes_read == 0)
-		return 0;
+		if (!(skip_zeroes && is_block_zeroed(buf, cmp_len))) {
+			checksum_block(buf, cmp_len, data->block_digest);
 
-	data->bytes += bytes_read;
+			ret = add_block_hash(&data->block_hashes,
+					     &data->nr_block_hashes,
+					     extoff + start,
+					     data->block_digest,
+					     flags);
+			if (ret)
+				break;
 
-	/* Handle partial read */
-	if (bytes_read < blocksize) {
-		/*
-		 * Don't want to store the len of each block, so
-		 * hash-tree makes the assumption that a partial block
-		 * is the last one.
-		 */
-		if (bytes_read + *off != data->file->size)
-			return -1;
-		partial = FILE_BLOCK_PARTIAL;
-	}
-	data->flags = hole = partial;
-	if (fc) {
-		unsigned int fieflags = 0;
-
-		ret = fiemap_iter_get_flags(fc, data->file, *off, &fieflags,
-					    &hole);
-		if (ret) {
-			fprintf(stderr,
-				"Fiemap error %d while scanning file "
-				"\"%s\": %s\n", ret, data->file->filename,
-				strerror(ret));
-
-			free(fc);
-			fc = *in_fc = NULL;
-		} else {
-			if (skip_zeroes && fieflags & FIEMAP_EXTENT_UNWRITTEN)
-				return 3;
-			if (hole)
-				return 3;
-			if (fieflags & FIEMAP_SKIP_FLAGS)
-				data->flags |= FILE_BLOCK_SKIP_COMPARE;
+			if (!v2_hashfile &&
+			    dbfile_cfg.extent_hash_src == EXTENT_HASH_SRC_DIGEST)
+				add_to_running_checksum(csum, digest_len,
+							data->block_digest);
 		}
+
+
+		start += cmp_len;
+		cmp_len = extlen - start;
+		if (cmp_len > blocksize)
+			cmp_len = blocksize;
 	}
 
-	if (skip_zeroes && is_block_zeroed(data->buf, data->bytes))
-		return 3;
+	if (!v2_hashfile &&
+	    dbfile_cfg.extent_hash_src == EXTENT_HASH_SRC_DATA)
+		add_to_running_checksum(csum, extlen,
+					(unsigned char *)data->buf);
+
+	assert(start == extlen);
+
+	return ret;
 
-	checksum_block(data->buf, data->bytes, data->digest);
-	if (data->flags & FILE_BLOCK_PARTIAL)
-		return 1;
-	return 2;
 }
 
-static void csum_whole_file_init(GMutex **mutex, void *location,
-				struct filerec *file, struct fiemap_ctxt **fc)
+static int csum_extent(struct csum_ctxt *data, uint64_t extent_off,
+		       unsigned int extent_len, int extent_flags,
+		       uint64_t *ret_total_bytes_read)
+{
+	int ret = 0;
+	int flags;
+	uint64_t total_bytes_read = 0;
+	struct running_checksum *csum;
+
+	csum = start_running_checksum();
+	if (!csum)
+		return -1;
+
+	while (total_bytes_read < extent_len) {
+		ssize_t bytes_read = 0;
+		unsigned int readlen = extent_len - total_bytes_read;
+		if (readlen > READ_BUF_LEN)
+			readlen = READ_BUF_LEN;
+
+		ret = pread(data->file->fd, data->buf, readlen, extent_off);
+		if (ret < 0) {
+			ret = errno;
+			fprintf(stderr, "Unable to read file %s: %s\n",
+				data->file->filename, strerror(ret));
+			return -ret;
+		}
+		if (ret == 0)
+			break;
+
+		bytes_read = ret;
+		total_bytes_read += bytes_read;
+		flags = xlate_extent_flags(extent_flags, bytes_read);
+
+		ret = csum_blocks(data, csum, extent_off, bytes_read, flags);
+		if (ret)
+			break;
+
+		extent_off += bytes_read;
+	}
+
+	finish_running_checksum(csum, data->digest);
+
+	*ret_total_bytes_read = total_bytes_read;
+	ret = (total_bytes_read == 0) ? 0 : 1; // handle overflow
+	return ret;
+}
+
+static void csum_whole_file_init(void *location, struct filerec *file,
+				 struct fiemap_ctxt **fc)
 {
 	static long long unsigned _cur_scan_files;
 	unsigned long long cur_scan_files;
-	*mutex = g_dataset_get_data(location, "mutex");
 
 	cur_scan_files = __atomic_add_fetch(&_cur_scan_files, 1, __ATOMIC_SEQ_CST);
 
@@ -676,32 +764,228 @@ static void csum_whole_file_init(GMutex **mutex, void *location,
 	}
 }
 
+
+/*
+ * Helper for csum_by_block/csum_by_extent.
+ * Return < 0 on error, 0 on success and 1 if we find an extent that should not
+ * be read.
+ */
+static int fiemap_helper(struct fiemap_ctxt *fc, struct filerec *file,
+			 uint64_t *poff, uint64_t *loff, uint32_t *len,
+			 unsigned int *flags)
+{
+	int ret;
+
+	ret = fiemap_iter_next_extent(fc, file, poff, loff, len, flags);
+	if (ret)
+		return ret;
+
+	if ((skip_zeroes && *flags & FIEMAP_EXTENT_UNWRITTEN) ||
+	    (*flags & FIEMAP_SKIP_FLAGS)) {
+		/*
+		 * Unritten or other extent we don't
+		 * want to read
+		 */
+		return 1;
+	}
+	return 0;
+}
+
+static int csum_by_block(struct csum_ctxt *ctxt, struct fiemap_ctxt *fc,
+			 struct block_csum **ret_block_hashes, int *ret_nb_hash)
+{
+	int ret;
+	uint64_t loff, poff, fieloff, bytes_read;
+	unsigned int fieflags, fielen;
+	struct filerec *file = ctxt->file;
+	struct block_csum *block_hashes;
+
+	block_hashes = malloc(sizeof(struct block_csum));
+	if (block_hashes == NULL)
+		return ENOMEM;
+
+	ctxt->block_hashes = block_hashes;
+        loff = fieloff = fielen = 0;
+	fieflags = 0;
+	while (loff < file->size) {
+		if (fc && loff >= (fieloff + fielen)) {
+			ret = fiemap_helper(fc, file, &poff, &fieloff, &fielen,
+					    &fieflags);
+			if (ret < 0)
+				return ret;
+			if (ret == 1) {
+				loff = fieloff + fielen;
+				continue;
+			}
+			loff = fieloff;
+			continue;
+		}
+
+//		printf("loff %"PRIu64"\n", loff);
+		ret = csum_extent(ctxt, loff, blocksize, fieflags, &bytes_read);
+		if (ret == 0) /* EOF */
+			break;
+
+		if (ret < 0) /* Err */
+			return ret;
+
+		if (bytes_read < blocksize && bytes_read + loff != file->size) {
+			/*
+			 * Don't want to store the len of each
+			 * block, so hash-tree makes the
+			 * assumption that a partial block is
+			 * the last one.
+			 */
+			return -1;
+		}
+		loff += blocksize;
+		if (bytes_read < blocksize) {
+			/* Partial read, don't get any more blocks */
+			break;
+		}
+	}
+	ret = 0;
+	ctxt->blocks_recorded = ctxt->nr_block_hashes;
+	*ret_nb_hash = ctxt->nr_block_hashes;
+	*ret_block_hashes = ctxt->block_hashes;
+
+	return ret;
+}
+
+static int get_file_extent_count(struct filerec *file, uint32_t *count)
+{
+	struct fiemap fiemap;
+	int err;
+
+	memset(&fiemap, 0, sizeof(fiemap));
+	fiemap.fm_length = ~0ULL;
+	err = ioctl(file->fd, FS_IOC_FIEMAP, (unsigned long) &fiemap);
+	if (err < 0)
+		return errno;
+
+	dprintf("Got %u extent for file\n", fiemap.fm_mapped_extents);
+	*count = fiemap.fm_mapped_extents;
+
+	return 0;
+}
+
+static int csum_by_extent(struct csum_ctxt *ctxt, struct fiemap_ctxt *fc,
+			  struct extent_csum **ret_extent_hashes,
+			  int *ret_nb_hash)
+{
+	uint64_t poff, loff, bytes_read;
+	uint32_t len, extents_count = 0;
+	int ret = 0;
+	unsigned int flags;
+	struct extent_csum *extent_hashes;
+	void *retp;
+	struct filerec *file = ctxt->file;
+	int nb_hash = 0;
+	struct block_csum *block_hashes;
+//	int nb_block_hash = 0;
+
+	ret = get_file_extent_count(file, &extents_count);
+	if (ret)
+		return ret;
+
+	extent_hashes = calloc(extents_count, sizeof(struct extent_csum));
+	if (extent_hashes == NULL)
+		return ENOMEM;
+
+	block_hashes = malloc(sizeof(struct block_csum));
+	if (block_hashes == NULL) {
+		free(extent_hashes);
+		return ENOMEM;
+	}
+
+	/* We require fiemap to do dedupe by extent. */
+	if (fc == NULL) {
+		free(extent_hashes);
+		free(block_hashes);
+		return ENOMEM;
+	}
+
+	ctxt->block_hashes = block_hashes;
+
+	flags = 0;
+	while (!(flags & FIEMAP_EXTENT_LAST)) {
+		ret = fiemap_helper(fc, file, &poff, &loff, &len, &flags);
+		if (ret < 0) {
+			fprintf(stderr, "Error %d from fiemap_helper()\n", ret);
+			goto out;
+		}
+
+		if (ret == 1) {
+			/* Skip reading this extent */
+			continue;
+		}
+
+		ret = csum_extent(ctxt, loff, len, flags, &bytes_read);
+		if (ret == 0) /* EOF */
+			break;
+
+		if (ret < 0)  /* Err */ {
+			fprintf(stderr, "Error %d from csum_extent()\n", ret);
+			goto out;
+		}
+
+		if ((nb_hash + 1) > extents_count) {
+			retp = realloc(extent_hashes,
+				       sizeof(struct extent_csum) * (nb_hash + 1));
+			if (!retp) {
+				ret = ENOMEM;
+				goto out;
+			}
+			extent_hashes = retp;
+		}
+
+//		printf("loff %"PRIu64" ret %d len %u flags 0x%x\n",
+//		       loff, ret, len, flags);
+		extent_hashes[nb_hash].loff = loff;
+		extent_hashes[nb_hash].poff = poff;
+		/* XXX: put len or actual read length in here? */
+//		extent_hashes[nb_hash].len = len;
+		extent_hashes[nb_hash].len = bytes_read;
+		extent_hashes[nb_hash].flags = flags;
+		memcpy(extent_hashes[nb_hash].digest, ctxt->digest,
+		       DIGEST_LEN_MAX);
+		nb_hash++;
+
+		ctxt->blocks_recorded += (bytes_read + (blocksize - 1))/blocksize;
+		if (bytes_read < len) {
+			/* Partial read, don't get any more blocks */
+			break;
+		}
+	}
+
+	ret = 0;
+	*ret_nb_hash = nb_hash;
+	*ret_extent_hashes = extent_hashes;
+out:
+	if (ret && extent_hashes)
+		free(extent_hashes);
+
+	return ret;
+}
+
 static void csum_whole_file(struct filerec *file,
 			    struct thread_params *params)
 {
-	uint64_t off = 0;
 	int ret = 0;
-	struct fiemap_ctxt *fc = NULL;
-	struct csum_block curr_block;
-	struct sqlite3 *db = NULL;
-
-	curr_block.buf = malloc(blocksize);
-	assert(curr_block.buf != NULL);
-	curr_block.file = file;
-	curr_block.bytes = 0;
-
-	struct block *hashes = malloc(sizeof(struct block));
-	void *retp;
 	int nb_hash = 0;
+	struct fiemap_ctxt *fc = NULL;
+	struct csum_ctxt csum_ctxt;
+	struct sqlite3 *db = NULL;
+	struct extent_csum *extent_hashes = NULL;
+	struct block_csum *block_hashes = NULL;
+	const char *errfunc = "(unknown)";
 
-	GMutex *mutex;
+	memset(&csum_ctxt, 0, sizeof(csum_ctxt));
+	csum_ctxt.buf = calloc(1, READ_BUF_LEN);
+	assert(csum_ctxt.buf != NULL);
+	csum_ctxt.file = file;
 
-	csum_whole_file_init(&mutex, params, file, &fc);
-
-	if (hashes == NULL) {
-		ret = ENOMEM;
-		goto err_noclose;
-	}
+	csum_whole_file_init(params, file, &fc);
 
 	db = dbfile_get_handle();
 	if (!db)
@@ -711,43 +995,18 @@ static void csum_whole_file(struct filerec *file,
 	if (ret)
 		goto err_noclose;
 
-	while (1) {
-		ret = csum_next_block(&curr_block, &off, &fc);
-		if (ret == 0) /* EOF */
-			break;
-
-		if (ret == -1) /* Err */
-			goto err;
-
-		if (ret == 3) { /* Skip block */
-			off += curr_block.bytes;
-			curr_block.bytes = 0;
-			continue;
-		}
-
-
-		retp = realloc(hashes, sizeof(struct block) * (nb_hash + 1));
-		if (!retp) {
-			ret = ENOMEM;
-			goto err;
-		}
-		hashes = retp;
-
-		hashes[nb_hash].loff = off;
-		hashes[nb_hash].flags = curr_block.flags;
-		memcpy(hashes[nb_hash].digest, curr_block.digest,
-					DIGEST_LEN_MAX);
-		nb_hash++;
-
-		if (ret == 1) /* Partial read, don't get any more blocks */
-			break;
-
-		off += curr_block.bytes;
-		curr_block.bytes = 0;
+	if (v2_hashfile) {
+		errfunc = "csum_by_block";
+		ret = csum_by_block(&csum_ctxt, fc, &block_hashes, &nb_hash);
+	} else {
+		errfunc = "csum_by_extent";
+		ret = csum_by_extent(&csum_ctxt, fc, &extent_hashes, &nb_hash);
 	}
+	if (ret)
+		goto err;
 
 	g_mutex_lock(&io_mutex);
-	file->num_blocks = nb_hash;
+	file->num_blocks = csum_ctxt.blocks_recorded;
 	/* Make sure that we'll check this file on any future dedupe passes */
 	filerec_clear_deduped(file);
 	ret = dbfile_begin_trans(db);
@@ -756,16 +1015,34 @@ static void csum_whole_file(struct filerec *file,
 		goto err;
 	}
 
-	ret = dbfile_write_file_info(db, file);
+	ret = dbfile_store_file_info(db, file);
 	if (ret) {
 		g_mutex_unlock(&io_mutex);
 		goto err;
 	}
 
-	ret = dbfile_write_hashes(db, file, nb_hash, hashes);
-	if (ret) {
-		g_mutex_unlock(&io_mutex);
-		goto err;
+	if (v2_hashfile) {
+		ret = dbfile_store_block_hashes(db, params->dbfile_cfg, file,
+						nb_hash, block_hashes);
+		if (ret) {
+			g_mutex_unlock(&io_mutex);
+			goto err;
+		}
+	} else {
+		ret = dbfile_store_block_hashes(db, params->dbfile_cfg, file,
+						csum_ctxt.nr_block_hashes,
+						csum_ctxt.block_hashes);
+		if (ret) {
+			g_mutex_unlock(&io_mutex);
+			goto err;
+		}
+
+		ret = dbfile_store_extent_hashes(db, params->dbfile_cfg, file,
+						 nb_hash, extent_hashes);
+		if (ret) {
+			g_mutex_unlock(&io_mutex);
+			goto err;
+		}
 	}
 
 	ret = dbfile_commit_trans(db);
@@ -775,71 +1052,83 @@ static void csum_whole_file(struct filerec *file,
 	}
 	g_mutex_unlock(&io_mutex);
 
-	g_mutex_lock(mutex);
+	g_mutex_lock(&params->mutex);
 	params->num_files++;
 	params->num_hashes += nb_hash;
-	g_mutex_unlock(mutex);
+	g_mutex_unlock(&params->mutex);
 
 	file->flags &= ~(FILEREC_NEEDS_SCAN|FILEREC_UPDATE_DB);
-	/* Set 'IN_DB' flag *after* we call dbfile_write_hashes() */
+	/* Set 'IN_DB' flag *after* we call dbfile_store_hashes() */
 	file->flags |= FILEREC_IN_DB;
 
 	filerec_close(file);
-	free(curr_block.buf);
+	free(csum_ctxt.buf);
 	if (fc)
 		free(fc);
+	if (csum_ctxt.block_hashes)
+		free(csum_ctxt.block_hashes);
 
-	free(hashes);
+	free(extent_hashes);
 	return;
 
 err:
 	filerec_close(file);
 err_noclose:
-	free(curr_block.buf);
-	if (hashes)
-		free(hashes);
+	free(csum_ctxt.buf);
+	if (extent_hashes)
+		free(extent_hashes);
+	if (block_hashes)
+		free(block_hashes);
+	if (csum_ctxt.block_hashes)
+		free(csum_ctxt.block_hashes);
 	if (fc)
 		free(fc);
 
 	fprintf(
 		stderr,
-		"Skipping file due to error %d (%s), %s\n",
+		"Skipping file due to error %d from function %s (%s), %s\n",
 		ret,
+		errfunc,
 		strerror(ret),
 		file->filename);
 
-	g_mutex_lock(mutex);
+	g_mutex_lock(&params->mutex);
 	/*
 	 * filerec_free will remove from the filerec tree keep it
 	 * under tree_mutex until we have a need for real locking in
 	 * filerec.c
 	 */
 	filerec_free(file);
-	g_mutex_unlock(mutex);
+	g_mutex_unlock(&params->mutex);
 
 	return;
 }
 
-int populate_tree()
+int populate_tree(struct dbfile_config *cfg)
 {
-	GMutex mutex;
 	GThreadPool *pool;
-	struct thread_params params = { 0, 0, };
+	struct thread_params params;
+	int ret = 0;
+
+	g_mutex_init(&params.mutex);
+	params.dbfile_cfg = cfg;
+	params.num_files = params.num_hashes = 0;
 
 	leading_spaces = num_digits(files_to_scan);
 
 	if (files_to_scan) {
-		pool = setup_pool(&params, &mutex, csum_whole_file);
-		if (!pool)
-			return ENOMEM;
+		pool = setup_pool(&params, csum_whole_file);
+		if (!pool) {
+			ret = ENOMEM;
+			goto out;
+		}
 
 		run_pool(pool);
 
-		qprintf("Total files:  %d\n", params.num_files);
-		qprintf("Total hashes: %d\n", params.num_hashes);
-
-		g_dataset_destroy(&params);
+		printf("Total files:  %d\n", params.num_files);
+		qprintf("Total extent hashes: %d\n", params.num_hashes);
 	}
-
-	return 0;
+out:
+	g_mutex_clear(&params.mutex);
+	return ret;
 }
