@@ -1,6 +1,8 @@
 #ifndef	__FILE_SCAN_H__
 #define	__FILE_SCAN_H__
 
+#include "list.h"
+
 /* from duperemove.c */
 extern int run_dedupe;
 extern int one_file_system;
@@ -12,7 +14,7 @@ extern int skip_zeroes;
 /*
  * Returns nonzero on fatal errors only
  */
-int add_file(const char *name, int dirfd);
+int add_file(const char *name);
 /*
  * Add from a db record. We still stat as before:
  *
@@ -31,13 +33,27 @@ void fs_set_onefs(dev_t dev, uint64_t fsid);
 dev_t fs_onefs_dev(void);
 uint64_t fs_onefs_id(void);
 
-int populate_tree();
+struct dbfile_config;
+int populate_tree(struct dbfile_config *cfg);
 
 /* For dbfile.c */
-struct block {
+struct block_csum {
 	uint64_t	loff;
 	unsigned int	flags;
 	unsigned char	digest[DIGEST_LEN_MAX];
 };
 
+struct extent_csum {
+	uint64_t	loff;
+	uint64_t	poff;
+	uint32_t	len;
+	unsigned int	flags;
+	unsigned char	digest[DIGEST_LEN_MAX];
+};
+
+struct exclude_file {
+	char *pattern;
+	struct list_head list;
+};
+
 #endif	/* __FILE_SCAN_H__ */
