@@ -7,5 +7,7 @@ extern int do_lookup_extents;
 extern unsigned int blocksize;
 
 int find_all_dupes(struct hash_tree *tree, struct results_tree *res);
+int find_additional_dedupe(struct hash_tree *hashes,
+			   struct results_tree *extents);
 
 #endif	/* __FIND_DUPES_H__ */
