@@ -41,7 +41,6 @@
 
 #include "run_dedupe.h"
 
-extern int block_dedupe;
 extern int dedupe_same_file;
 extern int fiemap_during_dedupe;
 
@@ -59,14 +58,11 @@ void print_dupes_table(struct results_tree *res)
 	struct dupe_extents *dext;
 	struct extent *extent;
 
-	if (quiet)
-		return;
-
 	printf("Simple read and compare of file data found %u instances of "
 	       "extents that might benefit from deduplication.\n",
 	       res->num_dupes);
 
-	if (res->num_dupes == 0)
+	if (quiet || res->num_dupes == 0)
 		return;
 
 	while (1) {
@@ -117,7 +113,7 @@ static void process_dedupe_results(struct dedupe_ctxt *ctxt,
 		    (target_status == -EINVAL && f == ctxt->ioctl_file))
 			continue;
 
-		if (target_status == BTRFS_SAME_DATA_DIFFERS)
+		if (target_status == FILE_DEDUPE_RANGE_DIFFERS)
 			status_str = "data changed";
 		else if (target_status < 0)
 			status_str = strerror(-target_status);
@@ -128,30 +124,32 @@ static void process_dedupe_results(struct dedupe_ctxt *ctxt,
 	}
 }
 
+#if 0
 static void get_extent_info(struct dupe_extents *dext)
 {
-	int ret = 0;
-	struct extent *extent;
-	struct filerec *file;
+       int ret = 0;
+       struct extent *extent;
+       struct filerec *file;
 
-	list_for_each_entry(extent, &dext->de_extents, e_list) {
-		file = extent->e_file;
+       list_for_each_entry(extent, &dext->de_extents, e_list) {
+               file = extent->e_file;
 
-		if (filerec_open(file, 0))
-			continue;
+               if (filerec_open(file, 0))
+                       continue;
 
-		extent_shared_bytes(extent) = 0;
-		ret = filerec_count_shared(file, extent->e_loff, dext->de_len,
-					   &extent_shared_bytes(extent),
-					   &extent_poff(extent),
-					   &extent_plen(extent));
-		if (ret) {
-			fprintf(stderr, "%s: fiemap error %d: %s\n",
-				extent->e_file->filename, ret, strerror(ret));
-		}
-		filerec_close(file);
-	}
+               extent_shared_bytes(extent) = 0;
+               ret = filerec_count_shared(file, extent->e_loff, dext->de_len,
+                                          &extent_shared_bytes(extent),
+                                          &extent_poff(extent),
+                                          &extent_plen(extent));
+               if (ret) {
+                       fprintf(stderr, "%s: fiemap error %d: %s\n",
+                               extent->e_file->filename, ret, strerror(ret));
+               }
+               filerec_close(file);
+       }
 }
+#endif
 
 static void add_shared_extents(struct dupe_extents *dext, uint64_t *shared)
 {
@@ -161,6 +159,36 @@ static void add_shared_extents(struct dupe_extents *dext, uint64_t *shared)
 		*shared += extent_shared_bytes(extent);
 }
 
+/*
+ * Fiemap the file and get our post-dedupe extent state.
+ *
+ * XXX: We're running fiemap too much for this. At the least we should
+ * shoot for one fiemap call(s) per filerec.
+ */
+static void add_shared_extents_post(struct dupe_extents *dext, uint64_t *shared)
+{
+	int ret;
+	uint64_t bytes = 0;
+	struct extent *extent;
+	struct filerec *file;
+
+	list_for_each_entry(extent, &dext->de_extents, e_list) {
+		file = extent->e_file;
+		ret = filerec_open(file, target_rw);
+		if (ret)
+			return;
+
+		ret = filerec_count_shared(file, extent->e_loff, dext->de_len,
+					   &bytes);
+
+		*shared += bytes;
+		filerec_close(file);
+
+		if (ret)
+			return;
+	}
+}
+
 static int disk_extent_grew(struct dupe_extents *dext, struct extent *extent)
 {
 	/*
@@ -199,9 +227,6 @@ static void clean_deduped(struct dupe_extents **ret_dext)
 	if (!dext || dext->de_num_dupes == 0)
 		return;
 
-	if (block_dedupe)
-		return;
-
 	outer = rb_first(&dext->de_extents_root);
 	while (outer) {
 		outer_extent = rb_entry(outer, struct extent, e_node);
@@ -301,10 +326,6 @@ static int dedupe_extent_list(struct dupe_extents *dext, uint64_t *fiemap_bytes,
 	 * the caller knows not to reference dext any more.
 	 */
 	if (fiemap_during_dedupe) {
-		ret = init_all_extent_dedupe_info(dext);
-		if (ret)
-			goto out;
-		get_extent_info(dext);
 		clean_deduped(&dext);
 		if (!dext) {
 			qprintf("[%p] Skipping - extents are already deduped.\n",
@@ -442,10 +463,9 @@ close_files:
 	abort_on(ctxt != NULL);
 	abort_on(!RB_EMPTY_ROOT(&open_files.root));
 
-	if (fiemap_during_dedupe) {
-		get_extent_info(dext);
-		add_shared_extents(dext, &shared_post);
-	}
+	if (fiemap_during_dedupe)
+		add_shared_extents_post(dext, &shared_post);
+
 	/*
 	 * It's entirely possible that some other process is
 	 * manipulating files underneath us. Take care not to
@@ -503,194 +523,13 @@ static int extent_dedupe_worker(struct dupe_extents *dext,
 	return 0;
 }
 
-struct block_dedupe_list
-{
-	unsigned char		bd_hash[DIGEST_LEN_MAX];
-	struct list_head	bd_block_list;
-	struct hash_tree	*bd_hash_tree;
-};
-static void free_bdl(struct block_dedupe_list *bdl);
-
-/*
- * tgt_file/tgt_off here are used only when we are asked not to dedupe
- * within the same file - see block_dedupe_nosame(). Otherwise this is
- * very straight-forward. We walk the dupe blocks list and insert into
- * the result tree then run that through dedupe_extent_list().
- */
-static int __block_dedupe(struct block_dedupe_list *bdl,
-			  struct results_tree *res,
-			  struct filerec *tgt_file,
-			  uint64_t tgt_off, uint64_t tgt_len,
-			  uint64_t *fiemap_bytes,
-			  uint64_t *kern_bytes, unsigned long long passno)
-{
-	int ret, one_old = 0;
-	struct dupe_extents *dext = NULL;
-	struct file_block *block;
-
-	if (tgt_file) {
-		/* Insert this first so it gets picked as target. */
-		ret = insert_one_result(res, bdl->bd_hash, tgt_file, tgt_off,
-					tgt_len);
-		if (ret)
-			return ret;
-
-		if (filerec_deduped(tgt_file))
-			one_old = 1;
-	}
-
-	list_for_each_entry(block, &bdl->bd_block_list, b_list) {
-		if (block->b_file == tgt_file)
-				continue;
-
-		if (filerec_deduped(block->b_file)) {
-			/*
-			 * Take one old result to dedupe from. That
-			 * way we don't get into a situation where
-			 * we've thrown out all blocks except a single
-			 * rescanned one.
-			 */
-			if (one_old)
-				continue;
-			one_old = 1;
-		}
-
-		ret = insert_one_result(res, bdl->bd_hash, block->b_file,
-					block->b_loff, block_len(block));
-		if (ret)
-			goto out;
-	}
-
-	dext = rb_entry(rb_first(&res->root), struct dupe_extents, de_node);
-	abort_on(!dext);
-
-	if (dext->de_num_dupes >= 2) {
-		ret = dedupe_extent_list(dext, fiemap_bytes, kern_bytes,
-					 passno);
-		if (ret == DEDUPE_EXTENTS_CLEANED)
-			ret = 0;
-	}
-
-out:
-	dext = rb_entry(rb_first(&res->root), struct dupe_extents, de_node);
-	if (dext) {
-		g_mutex_lock(&mutex);
-		dupe_extents_free(dext, res);
-		g_mutex_unlock(&mutex);
-	}
-
-	return ret;
-}
-
-/*
- * Newer kernels (Linux v4.2+) support dedupe of the target file
- * (dedupe with a file). We don't have a way of testing that yet so
- * for now assume we can't dedupe with the same file. Extent dedupe
- * handles this by never comparing a file to itself. We don't have
- * that luxury.
- *
- * Walk the dupe list once until we find two files which are
- * different. This is dumb but easy on memory and probably fast
- * enough.
- */
-static void pick_target_files(struct block_dedupe_list *bdl,
-			      struct filerec **tgt1, uint64_t *loff1,
-			      uint64_t *tlen1, struct filerec **tgt2,
-			      uint64_t *loff2, uint64_t *tlen2)
-{
-	struct filerec *file1, *file2;
-	uint64_t off1, off2, len1, len2;
-	struct file_block *block;
-
-	file1 = file2 = NULL;
-	off1 = off2 = len1 = len2 = 0;
-	list_for_each_entry(block, &bdl->bd_block_list, b_list) {
-		if (!file1) {
-			file1 = block->b_file;
-			off1 = block->b_loff;
-			len1 = block_len(block);
-		} else if (file1 != block->b_file) {
-			file2 = block->b_file;
-			off2 = block->b_loff;
-			len2 = block_len(block);
-		}
-
-		if (file1 && file2)
-			break;
-	}
-
-	*tgt1 = file1;
-	*tgt2 = file2;
-	*loff1 = off1;
-	*loff2 = off2;
-	*tlen1 = len1;
-	*tlen2 = len2;
-	return;
-}
-
-/* for kernels without same file dedupe (< v4.2) */
-static int block_dedupe_nosame(struct block_dedupe_list *bdl,
-			       struct results_tree *res, uint64_t *fiemap_bytes,
-			       uint64_t *kern_bytes, unsigned long long passno)
-{
-	int ret;
-	struct filerec *tgt_file1, *tgt_file2;
-	uint64_t tgt_off1, tgt_off2, tgt_len1, tgt_len2;
-
-	tgt_off1 = tgt_off2 = 0;
-
-	pick_target_files(bdl, &tgt_file1, &tgt_off1, &tgt_len1,
-			  &tgt_file2, &tgt_off2, &tgt_len2);
-
-	if (!tgt_file2) {
-		/* Can't dedupe this in nosame mode */
-		if (verbose) {
-			printf("[%p] Can not dedupe hash ", g_thread_self());
-			debug_print_digest_short(stdout, bdl->bd_hash);
-			printf(" in nosame mode - all hashes belong to %s.\n",
-			       tgt_file1->filename);
-		}
-		return 0;
-	}
-
-	ret = __block_dedupe(bdl, res, tgt_file1, tgt_off1, tgt_len1,
-			     fiemap_bytes, kern_bytes, passno);
-	if (!ret)
-		ret = __block_dedupe(bdl, res, tgt_file2, tgt_off2, tgt_len2,
-				     fiemap_bytes, kern_bytes, passno);
-
-	return ret;
-}
-
-static int block_dedupe_worker(struct block_dedupe_list *bdl,
-			       uint64_t *fiemap_bytes, uint64_t *kern_bytes)
-{
-	int ret;
-	struct results_tree res;
-	unsigned long long passno =  __atomic_add_fetch(&curr_dedupe_pass, 1, __ATOMIC_SEQ_CST);
-
-	init_results_tree(&res);
-
-	if (!dedupe_same_file)
-		ret = block_dedupe_nosame(bdl, &res, fiemap_bytes,
-					  kern_bytes, passno);
-	else
-		ret = __block_dedupe(bdl, &res, NULL, 0, 0, fiemap_bytes,
-				     kern_bytes, passno);
-	free_bdl(bdl);
-	return ret;
-}
-
 static int dedupe_worker(void *priv, struct dedupe_counts *counts)
 {
 	int ret;
 	uint64_t fiemap_bytes = 0ULL;
 	uint64_t kern_bytes = 0ULL;
 
-	if (block_dedupe)
-		ret = block_dedupe_worker(priv, &fiemap_bytes, &kern_bytes);
-	else
-		ret = extent_dedupe_worker(priv, &fiemap_bytes, &kern_bytes);
+	ret = extent_dedupe_worker(priv, &fiemap_bytes, &kern_bytes);
 
 	g_mutex_lock(&dedupe_counts_mutex);
 	counts->fiemap_bytes += fiemap_bytes;
@@ -702,150 +541,6 @@ static int dedupe_worker(void *priv, struct dedupe_counts *counts)
 
 static GThreadPool *dedupe_pool = NULL;
 
-struct block_dedupe_list *alloc_bdl(struct hash_tree *tree,
-				    struct dupe_blocks_list *dups)
-{
-	struct block_dedupe_list *bdl = malloc(sizeof(*bdl));
-
-	if (bdl) {
-		bdl->bd_hash_tree = tree;
-		INIT_LIST_HEAD(&bdl->bd_block_list);
-		memcpy(&bdl->bd_hash, &dups->dl_hash, DIGEST_LEN_MAX);
-	}
-	return bdl;
-}
-
-static void free_bdl(struct block_dedupe_list *bdl)
-{
-	struct file_block *block, *tmp;
-
-	if (bdl) {
-		list_for_each_entry_safe(block, tmp, &bdl->bd_block_list, b_list) {
-			list_del_init(&block->b_list);
-			g_mutex_lock(&mutex);
-			remove_hashed_block(bdl->bd_hash_tree, block);
-			g_mutex_unlock(&mutex);
-
-		}
-		free(bdl);
-	}
-}
-
-static int push_bdl(struct block_dedupe_list *bdl)
-{
-	GError *err = NULL;
-
-	g_thread_pool_push(dedupe_pool, bdl, &err);
-	if (err) {
-		fprintf(stderr, "Fatal error while deduping: %s\n",
-			err->message);
-		g_error_free(err);
-		return 1;
-	}
-	return 0;
-}
-
-/*
- * Don't queue more than DEDUPE_MAX dedupes on any one thread. We want
- * this because large lists (in the multiple millions) can wind up
- * taking longer on one thread than all the other lists. This happens
- * because a common pattern is only one or two extremely large buckets
- * (think zeros) with the rest having a drastically smaller number of
- * duplicates.
- */
-#define	DEDUPE_MAX	(1000000)
-static int __push_blocks(struct hash_tree *hashes,
-			 struct dupe_blocks_list *dups)
-{
-	int i, ret = 0;
-	unsigned long long j;
-	struct block_dedupe_list *bdl;
-	struct file_block *block;
-
-	if (dups->dl_num_elem < DEDUPE_MAX) {
-		/* the easy way */
-		bdl = alloc_bdl(hashes, dups);
-		if (!bdl) {
-			fprintf(stderr, "Fatal: out of memory while deduping\n");
-			ret = ENOMEM;
-			goto out;
-		}
-
-		list_splice_init(&dups->dl_list, &bdl->bd_block_list);
-		if (push_bdl(bdl))
-			goto out;
-	} else {
-		unsigned long long smax = dups->dl_num_elem / io_threads;
-		if (verbose) {
-			printf("Hash ");
-			debug_print_digest_short(stdout, dups->dl_hash);
-			printf(" has %llu elements. It will be spread amongst "
-			       "all dedupe threads.\n", dups->dl_num_elem);
-		}
-
-		for (i = 0; i < io_threads; i++) {
-			j = 0;
-			list_for_each_entry(block, &dups->dl_list, b_list) {
-				j++;
-
-				/*
-				 * Give the last thread any leftovers
-				 * even if that goes above our soft
-				 * max.
-				 */
-				if (list_is_last(&block->b_list, &dups->dl_list)
-				    || (j == smax && i != (io_threads - 1))) {
-					bdl = alloc_bdl(hashes, dups);
-					if (!bdl) {
-						fprintf(stderr,
-							"Fatal: out of memory"
-							" while deduping\n");
-						ret = ENOMEM;
-						goto out;
-					}
-
-					list_cut_position(&bdl->bd_block_list,
-							  &dups->dl_list,
-							  &block->b_list);
-
-					if (push_bdl(bdl))
-						goto out;
-					bdl = NULL;
-
-					 __atomic_add_fetch(
-						&total_dedupe_passes, 1, __ATOMIC_SEQ_CST);
-					break;
-				}
-			}
-		}
-	}
-
-	bdl = NULL; /* freed by thread */
-out:
-	free_bdl(bdl);
-	return ret;
-}
-
-static int push_blocks(struct hash_tree *hashes)
-{
-	struct dupe_blocks_list *dups, *tmp;
-
-	list_for_each_entry_safe(dups, tmp, &hashes->size_list, dl_size_list) {
-		if (dups->dl_num_elem < 2)
-			continue;
-		/*
-		 * Do this here because we don't want one of the
-		 * worker threads removing it from the size list for
-		 * us when they free the dups structure.
-		 */
-		list_del_init(&dups->dl_size_list);
-
-		if (__push_blocks(hashes, dups))
-			return 1;
-	}
-	return 0;
-}
-
 /* Errors from this function are fatal. */
 static int push_extents(struct results_tree *res)
 {
@@ -885,15 +580,13 @@ void dedupe_results(struct results_tree *res, struct hash_tree *hashes)
 	struct dedupe_counts counts = { 0ULL, };
 	GError *err = NULL;
 
-	if (!block_dedupe) {
-		results_tree = res;
+	results_tree = res;
 
-		print_dupes_table(res);
+	print_dupes_table(res);
 
-		if (RB_EMPTY_ROOT(&res->root)) {
-			printf("Nothing to dedupe.\n");
-			return;
-		}
+	if (RB_EMPTY_ROOT(&res->root)) {
+		printf("Nothing to dedupe.\n");
+		return;
 	}
 
 	qprintf("Using %u threads for dedupe phase\n", io_threads);
@@ -907,16 +600,9 @@ void dedupe_results(struct results_tree *res, struct hash_tree *hashes)
 		return;
 	}
 
-	if (block_dedupe) {
-		total_dedupe_passes = hashes->num_hashes;
-		leading_spaces = num_digits(total_dedupe_passes);
-		ret = push_blocks(hashes);
-	} else {
-		total_dedupe_passes = res->num_dupes;
-		leading_spaces = num_digits(total_dedupe_passes);
-		ret = push_extents(res);
-	}
-
+	total_dedupe_passes = res->num_dupes;
+	leading_spaces = num_digits(total_dedupe_passes);
+	ret = push_extents(res);
 	if (ret) {
 		fprintf(stderr, "Fatal error while deduping: %s\n",
 			err->message);
